/* Snowlog - An Apache logfile browser/analyzer.
 * 
 * Copyright 2004 Oliver Feiler <kiza@kcore.de>
 * http://snowlog.kcore.de/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as   
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <pcre.h>

#ifndef CONFIG_H
#define CONFIG_H

struct logentry;

struct host {
	char * ipaddr;
	char * hostname;
	char * first_referer;
	char * smart_first_referer;
	struct logentry ** entries;
	int num_entries;
	unsigned long long total_bytes;
};

struct logentry {
	int httpversion;
	int httpstatus;
	unsigned long long content_length;
	int timestamp;
	char const * method;
	char const * request;
	char const * mime_type;
	char const * referer;
	char const * useragent;
	char const * vhost;
	char const * authname;
	char const * loghint;
	/* Magic stuff */
	int isbot;					/* Try to figure out if a certain user agent
								   is a bot and set to 1 if yes.
								   UI -> hide all bot requests. */
};

typedef enum {
	FILTER_NONE,
	FILTER_HTTPSTATUS,
	FILTER_CONTENT_LENGTH,
	FILTER_METHOD,
	FILTER_REQUEST,
	FILTER_MIME_TYPE,
	FILTER_REFERER,
	FILTER_USERAGENT,
	FILTER_VHOST,
	FILTER_AUTHNAME,
	FILTER_LOGHINT,
	FILTER_NOBOTS
} filter_type;

typedef enum {
	FILTER_EQUALS,
	FILTER_GREATER_THAN,
	FILTER_LESS_THAN,
	FILTER_ISNOT
} filter_operator;

struct displayitem {
	int skip;
	int indent;
	int type;
	int num_columns;
	const char **columns;
	void *object;
};

struct regexp {
	char *regexp;
	char *substitution;
	pcre *compiled;
	pcre_extra *studied;
	int regexp_pattern_count;
};

struct displaystats {
	unsigned long long count_hosts;
	unsigned long long count_requests;
	unsigned long long count_bytes;
};

struct filter_rule {
	filter_type type;
	filter_operator operator;
	void *filter;
};

struct filter {
	char *name;
	int num_filters;
	struct filter_rule ** filter_rules;
};

#endif
