/* Snowlog - An Apache logfile browser/analyzer.
 * 
 * Copyright 2004 Oliver Feiler <kiza@kcore.de>
 * http://snowlog.kcore.de/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as   
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
   
#include <ncurses.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "config.h"

#include "snowtk.h"
#include "main.h"
#include "internals.h"
#include "regexp.h"

extern struct host **hosts;
extern int count_requests, count_hosts;
extern struct regexp ** useragents;
extern int useragents_count;
extern struct regexp ** referers;
extern int referers_count;
extern struct filter ** filters;
extern int count_filters;

extern int first_date;
extern int last_date;

enum {
	SNOW_COLOR_ENABLE,
	SNOW_COLOR_DISABLE
};

void RequestDetail (struct displayitem * item);
void UIColorToggle (const char * mime_type, int toggle);
void UIHTTPStatColorToggle (int httpstatus, int toggle);
int UIFilter (void);


void UIMainInterface (void) {
	int uiinput, i, j, ypos, itemnum, maxchars;
	int forceredraw = 0;
	int highlightnum = 1;
	int offset = 0;
	int maxitems;					/* Number of displayitems */
	int highlighted = 0;
	int is_highlighted;
	int filter_to_use = 0;
	int old_filter = 0;
	int typeahead = 0;
	struct displaystats *stats;
	struct displayitem **items;
	struct displayitem *cur;
	char tmp[512];
	char bytestring[64];
	/* Type ahead find */
	char *search = NULL;
	int searchlen = 0;
	int old_highlighted = 0;
	int old_offset = 0;
	int found = 0;
	int skip = 0;
	int skipper = 0;
	
	clear();
	
	stats = NewDisplayStatsStruct();
	items = DisplayItemsFromHosts(hosts, stats, filter_to_use);
	maxitems = stats->count_hosts+stats->count_requests;
	
	while (1) {
		if (old_filter != filter_to_use) {
			/* Reset display position or crash. */
			highlighted = 0;
			offset = 0;
			FreeDisplayList(items, maxitems);
			items = DisplayItemsFromHosts(hosts, stats, filter_to_use);
			maxitems = stats->count_hosts+stats->count_requests;
			old_filter = filter_to_use;
		}
		SnowClearScreen(&forceredraw);
		
		if (filter_to_use)
			UISupportDrawHeader(filters[filter_to_use-1]->name);
		else
			UISupportDrawHeader(NULL);
		
		ypos = 1;
		itemnum = 1;
		
		if (typeahead) {
			/* Search from topmost item */
			highlighted = 0;
			offset = 0;
			found = 0;
			skipper = 0;
			
			if (searchlen > 0) {
				for (i = 0; i < maxitems; i++) {
					if ((strstr(items[i]->columns[0], search) != NULL) ||
						(strstr(items[i]->columns[1], search) != NULL)) {
						found = 1;
						if (skip >= 1) {
							if (skipper == skip) {
								break;
							} else {
								skipper++;
								/* We need to increase counter also, because we
								don't reach the end of the loop after continue!. */
								highlighted++;
								offset++;
								continue;
							}
						}
						break;
					}
					highlighted++;
					offset++;
				}
			}
			if (!found) {
				highlighted = old_highlighted;
				offset = old_offset;
			}
		}
		
		for (j = 0; j < LINES-8; j++) {
			if ((items == NULL) || (maxitems == 0))
				break;
			
			cur = items[j+offset];
			
			/* Seems to be invoked from type ahead */
			if (offset == maxitems) {
				UIStatus("FUCKED!!!", 2, 0, LINES-1);
				snprintf(tmp, sizeof(tmp), "Oh no, not again! The annoying offset==maxitems bug. :(\n\nmaxitems: %d\nhighlighted (old): %d (%d)\noffset (old): %d (%d)\nskip: %d\nskipper: %d\nj: %d\n\n",
											maxitems,
											highlighted,
											old_highlighted,
											offset,
											old_offset,
											skip,
											skipper,
											j);
				MainQuit(tmp);
			}
			
			SnowClearLine(ypos);
			if (j+offset == highlighted)
				is_highlighted = 1;
			else
				is_highlighted = 0;
			
			if (cur->type == ENTRY_TYPE_HOST)
				attron(WA_BOLD);
			
			if (is_highlighted) {
				highlightnum = itemnum;
				attron (WA_REVERSE);
				for (i = 0; i < COLS; i++)
					mvaddch (ypos, i, ' ');
			}
			
			if (cur->type == ENTRY_TYPE_HOST) {
				maxchars = COLS-17-9;
				mvaddnstr (ypos, 1, cur->columns[0], 16);
				mvaddnstr (ypos, 17, cur->columns[1], maxchars);
				if (strlen(cur->columns[1]) > maxchars)
					mvaddstr (ypos, 16+maxchars+1, "...");
			} else if (cur->type == ENTRY_TYPE_REQUEST) {
				maxchars = COLS-10;
				if (!is_highlighted)
					UIColorToggle(((struct logentry *)(cur->object))->mime_type, SNOW_COLOR_ENABLE);
				mvaddnstr (ypos, 3, cur->columns[0], maxchars);
				if (strlen(cur->columns[0]) > maxchars)
					mvaddstr (ypos, maxchars+1, "...");
				if (!is_highlighted)
					UIColorToggle(((struct logentry *)(cur->object))->mime_type, SNOW_COLOR_DISABLE);

				
				/* Coloured HTTP status */
				if (!is_highlighted)
					UIHTTPStatColorToggle (((struct logentry *)(cur->object))->httpstatus,
										SNOW_COLOR_ENABLE);
				mvaddstr (ypos, COLS-5, cur->columns[1]);
				if (!is_highlighted)
					UIHTTPStatColorToggle (((struct logentry *)(cur->object))->httpstatus,
										SNOW_COLOR_DISABLE);
			}
			
			ypos++;
			if (is_highlighted)
				attroff (WA_REVERSE);
			
			if (cur->type == ENTRY_TYPE_HOST)
				attroff(WA_BOLD);
						
			itemnum++;
			if (j+offset+1 >= maxitems)
				break;
		}
		
		UISupportDrawBox(0, LINES-7, COLS, LINES-7);
		attron(WA_REVERSE);
		mvprintw(LINES-7, 1, "Line %d/%d", highlighted+1, maxitems);
		attroff(WA_REVERSE);

		if (items != NULL) {
			if (items[highlighted] != NULL)
				RequestDetail(items[highlighted]);
		}
		
		if (typeahead) {
			snprintf (tmp, sizeof(tmp), "-> %s", search);
			UIStatus (tmp, 0, 0, LINES-1);
		} else {
			StringWithSmartBytes (bytestring, sizeof(bytestring), stats->count_bytes);
			snprintf(tmp, sizeof(tmp),
				"%llu requests from %llu hosts (%s). (H)elp.", stats->count_requests, stats->count_hosts, bytestring);
			UIStatus(tmp, 0, 0, LINES-1);
		}
		uiinput = getch();
		
		if (typeahead) {
			/* Only match real characters. */
			if ((uiinput >= 32) && (uiinput <= 126)) {
				search[searchlen] = uiinput;
				search[searchlen+1] = 0;
				searchlen++;
				search = realloc (search, searchlen+2);
			} else if ((uiinput == 127) || (uiinput == 263)) {
				/* Don't let searchlen go below 0! */
				if (searchlen > 0) {
					searchlen--;
					search[searchlen] = 0;
					search = realloc (search, searchlen+2);
				} else {
					/* And if it does, stop typeahead. */
					typeahead = 0;
					free (search);
				}
			} else if ((uiinput == '\n') ||
						(uiinput == KEY_UP) || (uiinput == KEY_DOWN) ||
						(uiinput == KEY_LEFT) || (uiinput == KEY_RIGHT)) {
				/* Stop typeahead */
				typeahead = 0;
				free(search);
			}
		} else {
			if (uiinput == 'q') {
				MainQuit(NULL);
			} else if (uiinput == 12) {	/* Redraw screen on ^L */
				forceredraw = 1;
			} else if ((uiinput == KEY_DOWN) ||	/* Line by line */
				(uiinput == 'n')) {
				if (highlighted < maxitems-1) {
					highlighted++;
					if (highlightnum+1 > LINES-8)
						offset++;
				}
			} else if ((uiinput == KEY_UP) ||
				(uiinput == 'p')) {
				if (highlighted > 0) {
					highlighted--;
					if (highlightnum-1 < 1)
						offset--;
				}
			} else if ((uiinput == KEY_NPAGE) || (uiinput == 32)) {	/* Page by page */
				for (i = 0; i <= LINES-9; i++) {
					if (highlighted < maxitems-1) {
						highlighted++;
						offset++;
					} else
						break;
				}
			} else if ((uiinput == KEY_PPAGE) || (uiinput == 'b')) {
				for (i = 0; i <= LINES-9; i++) {
					if (highlighted > 0) {
						highlighted--;
						if (offset > 0)
							offset--;
					} else
						break;
				}
			} else if ((uiinput == KEY_HOME) || (uiinput == '<')) {	/* Top/bottom */
				highlighted = 0;
				offset = 0;
			} else if ((uiinput == KEY_END) || (uiinput == '>')) {
				highlighted = maxitems-1;
				offset = maxitems-1;
			} else if ((uiinput == 'N') || (uiinput == KEY_RIGHT)) {	/* Jump to next/previous host */
				for (;;) {
					if (highlighted < maxitems-1) {
						highlighted++;
						highlightnum++;
						offset++;
					}
					if (highlighted+1 >= maxitems-1)
						break;
					if (items[highlighted]->type == ENTRY_TYPE_HOST)
						break;
				}
			} else if ((uiinput == 'P') || (uiinput == KEY_LEFT)) {
				for (;;) {
					if (highlighted > 0) {
						highlighted--;
						highlightnum--;
						if (offset > 0)
							offset--;
					}
					if (highlighted-1 < 1)
						break;
					if (items[highlighted]->type == ENTRY_TYPE_HOST)
						break;
				}
			} else if (uiinput == 'o') {	/* Open referer URL */
				if (items != NULL) {
					if (items[highlighted] != NULL) {
						if (items[highlighted]->type == ENTRY_TYPE_REQUEST)
							UISupportURLJump(((struct logentry *)(items[highlighted]->object))->referer);
						else if (items[highlighted]->type == ENTRY_TYPE_HOST)
							UISupportURLJump(((struct host *)(items[highlighted]->object))->first_referer);
					}
				}
			} else if (uiinput == 'O') {	/* Open current resource */
				if (items != NULL) {
					if (items[highlighted] != NULL) {
						if (items[highlighted]->type == ENTRY_TYPE_REQUEST) {
							snprintf(tmp, sizeof(tmp), "http://%s%s",
									((struct logentry *)(items[highlighted]->object))->vhost,
									((struct logentry *)(items[highlighted]->object))->request);
							UISupportURLJump(tmp);
						}
					}
				}
			} else if (uiinput == 'f') {
				filter_to_use = UIFilter();
			} else if ((uiinput == 'h') || (uiinput == 'H')) {
				Help();
			}
		}
	
		/* Type ahead find */
		if ((uiinput == 9) || (uiinput == '/')) {
			if (typeahead) {
				if (searchlen == 0) {
					typeahead = 0;
					/* Typeahead now off. */
					free (search);
					skip = 0;
					//highlighted = old_highlighted;
					//offset = old_offset;
				} else {
					found = 0;
					for (i = 0; i < maxitems; i++) {
						/* Substring match. */
						if ((strstr(items[i]->columns[0], search) != NULL) ||
							(strstr(items[i]->columns[1], search) != NULL))
							found++;
					}
					if (skip == found-1) { /* found-1 to avoid empty tab cycle. */
						skip = 0;
					} else if (found > 1) {
						/* If more than one match was found and user presses tab we will skip matches. */
						skip++;
					}
				}
			} else {
				typeahead = 1;
				/* Typeahead now on. */
				search = malloc (2);
				memset (search, 0, 2);
				searchlen = 0;
				skip = 0;
				/* Save these offsets before we start typeahead. */
				old_highlighted = highlighted;
				old_offset = offset;
			}
		}
	}
}


void RequestDetail (struct displayitem * item) {
	int maxchars;
	int ypos = LINES-6;
	char bytestring[64];
	char *timestring;
	time_t t;
	char tmp[256];
	char *uistring;
	struct logentry * current_entry = item->object;
	
	if (item->type == ENTRY_TYPE_REQUEST) {
		t = current_entry->timestamp;
		timestring = ctime (&t);
		
		/* Line 1 */
		mvaddnstr (ypos, 1, current_entry->method, 8);
		maxchars = COLS-16;
		mvaddnstr (ypos, 8+1, current_entry->request, maxchars);
		if (strlen(current_entry->request) > maxchars)
			mvaddstr (ypos, 6+maxchars+1, "...");
		UIHTTPStatColorToggle (current_entry->httpstatus, SNOW_COLOR_ENABLE);
		mvprintw (ypos, 9+1+(COLS-15), "%d", current_entry->httpstatus);
		UIHTTPStatColorToggle (current_entry->httpstatus, SNOW_COLOR_DISABLE);
		
		/* Line 2 */
		mvaddstr (ypos+1, 1, timestring);
		
		StringWithSmartBytes (bytestring, sizeof(bytestring), current_entry->content_length);
		snprintf (tmp, sizeof(tmp), "%s, %s",
				  current_entry->mime_type, bytestring);
		UIColorToggle (current_entry->mime_type, SNOW_COLOR_ENABLE);
		mvaddstr (ypos+1, COLS-strlen(tmp)-2, tmp);
		UIColorToggle (current_entry->mime_type, SNOW_COLOR_DISABLE);
		
		/* Line 3 */
		uistring = NewRegexpMatchedString(useragents, useragents_count,
										  current_entry->useragent);
		mvaddnstr (ypos+2, 1, uistring, COLS-2);
		free(uistring);
		
		/* Line 4 */
		mvaddstr (ypos+3, 1, current_entry->vhost);
		if (current_entry->authname &&
			(current_entry->authname[0] != '-') &&
			(current_entry->httpstatus != 401)) { /* Remote user (from auth; may be bogus if return status (%s) is 401) - Apache docs mod_log */
			snprintf (tmp, sizeof(tmp), "Authenticated: %s", current_entry->authname);
			attron(COLOR_PAIR(SNOW_COLOR_RED));
			mvaddstr (ypos+3, COLS-strlen(tmp)-2, tmp);
			attroff(COLOR_PAIR(SNOW_COLOR_RED));
		}
		
		/* Line 5 */
		maxchars = COLS-5;
		mvaddnstr (ypos+4, 1, current_entry->referer, COLS-2);
		if (strlen(current_entry->referer) > maxchars)
			mvaddstr (ypos+4, COLS-4, "...");
			
	} else if (item->type == ENTRY_TYPE_HOST) {
		/* Line 1*/
		StringWithSmartBytes (bytestring, sizeof(bytestring),
							  ((struct host *)(item->object))->total_bytes);
		mvprintw (ypos, 1, "Total number of requests: %d, %s",
				  ((struct host *)(item->object))->num_entries, bytestring);
		mvaddstr (ypos+1, 1, "First referer:");
		mvaddnstr (ypos+1, 2+strlen("First referer:"),
				  ((struct host*)(item->object))->first_referer,
				  COLS-strlen("Main referer:")-4);
	}
}

void UIColorToggle (const char * mime_type, int toggle) {
	int snow_color;
	
	if (strstr(mime_type, "image") != NULL)
		snow_color = SNOW_COLOR_MAGENTA;
	else if (strstr(mime_type, "application") != NULL)
		snow_color = SNOW_COLOR_GREEN;
	else if (strstr(mime_type, "video") != NULL)
		snow_color = SNOW_COLOR_CYAN;
	else if (strstr(mime_type, "audio") != NULL)
		snow_color = SNOW_COLOR_BLUE;
	else return;

	if (toggle == SNOW_COLOR_ENABLE) {
		attron(COLOR_PAIR(snow_color));
	} else if (toggle == SNOW_COLOR_DISABLE) {
		attroff(COLOR_PAIR(snow_color));
	}
}

void UIHTTPStatColorToggle (int httpstatus, int toggle) {
	int snow_color;
	
	if ((httpstatus >= 200) && (httpstatus < 300))
		snow_color = SNOW_COLOR_BLUE;
	else if ((httpstatus >= 300) && (httpstatus < 400))
		snow_color = SNOW_COLOR_ORANGE;
	else if (httpstatus >= 400)
		snow_color = SNOW_COLOR_RED;
	else return;
	
	if (toggle == SNOW_COLOR_ENABLE) {
		attron(COLOR_PAIR(snow_color));
	} else if (toggle == SNOW_COLOR_DISABLE) {
		attroff(COLOR_PAIR(snow_color));
	}
}

int UIFilter (void) {
	int i, j, ypos;
	int uiinput = 0;
	
	j = 0;
	while (uiinput != '\n') {
		ypos = 4;
		UISupportDrawBox((COLS/2)-30, ypos, (COLS/2)+30, ypos+count_filters+3);
		
		attron(WA_REVERSE);
		mvaddstr(ypos+1, (COLS/2)-(strlen("Select filter"))/2, "Select filter");
		ypos = 7;
		
		if (count_filters == 0)
			mvaddstr (ypos-1, (COLS/2)-28, "No filters defined");
		for (i = 0; i < count_filters; i++) {
			j = i+49;
			/* Fast forward to 'a' if we hit ASCII 58 ':' */
			if (j >= 58)
				j += 39;
			mvprintw (ypos+i, (COLS/2)-28, "%c. %s", j, filters[i]->name);
		}
		attroff(WA_REVERSE);
		
		refresh();
		uiinput = getch();
		
		/* Always return 0 if no filters exist! */
		if (count_filters < 1)
			return 0;
		
		if (uiinput <= j) {
			if ((uiinput >= 49) && (uiinput <= 57))
				return uiinput-48;
			if ((uiinput >= 97) && (uiinput <= 122))
				return uiinput-87;
		}
	}
	return 0;
}
